#!/usr/bin/env python3

from qgis.core import *
from qgis.gui import *
from qgis.PyQt.QtXml import QDomDocument, QDomElement
from qgis.PyQt.QtWidgets import *
from qgis.PyQt.QtCore import *
from qgis.PyQt.QtGui import *
import math


from qasync import QEventLoop, asyncClose, asyncSlot, QApplication
import asyncio

from utils import *
import communication
import data_utils
from data_utils import Trip, TripPoint, HistoryPoint, TripHistory, shape_indexer, dist, unzip_parse
import graph
import graph

import datetime
import json
import sys, os
import pprint
import gtfs

import math

import client_config
from client_config import get_communication


window_title_prefix = "PID realtime"

QgsApplication.setPrefixPath("/usr/", True)
qgs = QgsApplication([], True)
qgs.initQgis()

crs_wgs=QgsCoordinateReferenceSystem(4326)
crs_wgs_mercator=QgsCoordinateReferenceSystem(3857)
crs_transform = QgsCoordinateTransform(crs_wgs, crs_wgs_mercator, QgsProject.instance())

live_subscribers = set()
live_data = None
live_data_by_trip_id = None
live_dt = None

def error(s):
    eprint("ERROR:", s)

fetch_live_task: asyncio.Task = None

async def fetch_live():
    global live_data, live_dt, live_data_by_trip_id
    eprint("Hi from fetch_live")
    c = await get_communication()
    r = await c.get_last_data()
    live_data = await unzip_parse(r[1])
    live_data_by_trip_id = {x["properties"]["trip"]["gtfs"]["trip_id"]: x for x in live_data["features"]}
    live_dt = r[0]
    for f in live_subscribers:
        asyncio.create_task(f(live_dt, live_data, live_data_by_trip_id))

    while True:
        eprint("Start wait next")
        r = await c.wait_next_data(live_dt)
        eprint("Done wait next")
        live_data = await unzip_parse(r[1])
        live_data_by_trip_id = {x["properties"]["trip"]["gtfs"]["trip_id"]: x for x in live_data["features"]}
        live_dt = r[0]
        for f in live_subscribers:
            asyncio.create_task(f(live_dt, live_data, live_data_by_trip_id))


def spawn_or_remove_fetch_live_task():
    global live_data, live_dt, fetch_live_task
    eprint("spawn_or_remove_fetch_live_task")
    if fetch_live_task and len(live_subscribers) == 0:
        fetch_live_task.cancel()
        fetch_live_task = None
        live_data = None
        live_dt = None
    if not fetch_live_task and len(live_subscribers):
        fetch_live_task = asyncio.create_task(catch_all(fetch_live()))



class ClickableLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setStyleSheet ("color: blue; text-decoration: underline;");

    clicked = pyqtSignal()
    def mouseReleaseEvent(self, event):
        super(ClickableLabel, self).mouseReleaseEvent(event)
        if event.button() == Qt.LeftButton and event.pos() in self.rect():
            self.clicked.emit()

class DockWidgetWithClosedSignal(QDockWidget):
    closed = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def closeEvent(self, event):
        self.closed.emit()
        super().closeEvent(event)

def no_space(lay):
    lay.setSpacing(0)
    lay.setContentsMargins(0, 0, 0, 0)
    return lay





async def gtfs_default_data_getter(date, filename):
    c =  await get_communication()
    return await c.gtfs_get_file(date, filename)

gtfs.default_data_getter = gtfs_default_data_getter
data_utils.get_communication = get_communication
graph.get_communication = get_communication

def layer_updated(layer):
    prov = layer.dataProvider()
    layer.updateExtents()
    layer.triggerRepaint()
    prov.reloadData()


def tool_icon(icon_path, slot):
    button = QToolButton()
    button.setIcon(QIcon(icon_path));
    button.setIconSize(QSize(25, 25));
    return button


class TimeControllToolBar(QToolBar):
    def __init__(self, wnd):
        super().__init__("Time Controll")
        self.wnd = wnd

        self._time_label = QLabel()
        f = self._time_label.font()
        f.setPixelSize(22)
        self._time_label.setFont(f)
        self._time_label.setFixedWidth(250)
        self.addWidget(self._time_label)

        self._select_date_time = QAction(QIcon(":images/themes/default/mActionOpenTableEdited.svg"),"Set date and time")
                                         #wnd.set_date_time)
        self._select_date_time.triggered.connect(wnd.set_date_time)
        self.addAction(self._select_date_time)

        self._prev_date_time = QAction(QIcon(":images/themes/default/temporal_navigation/previous.svg"), "Previous captured")
        self._prev_date_time.triggered.connect(wnd.prev_data)
        self.addAction(self._prev_date_time)

        self._next_date_time = QAction(QIcon(":images/themes/default/temporal_navigation/next.svg"), "Next captured")
        self._next_date_time.triggered.connect(wnd.next_data)
        self.addAction(self._next_date_time)


        self._start_live = QAction(QIcon(":images/themes/default/temporal_navigation/forward.svg"), "Watch live state")
        self._start_live.triggered.connect(wnd.toogle_live)
        self.addAction(self._start_live)


        wnd.shown_data_changed_signal.connect(self.time_changed)

    def time_changed(self):
        self._time_label.setText(format_dt_to_user(self.wnd.current_data_capture_time) if self.wnd.current_data_capture_time else "")


class FilterWidget(QWidget):
    filter_changed = pyqtSignal()

    state_position_types = ["before_track", "before_track_delayed", "after_track", "on_track", "at_stop", "off_track"]

    def __init__(self, wnd):
        super().__init__()
        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)

        self._state_positions = {}
        for state_position_type in self.state_position_types:
            self._state_positions[state_position_type] = QCheckBox(state_position_type.replace("_", " "))
            self._state_positions[state_position_type].stateChanged.connect(self.ui_changed)
            self._layout.addWidget(self._state_positions[state_position_type])
        self._state_positions["on_track"].setChecked(True)
        self._state_positions["at_stop"].setChecked(True)
        self._state_positions["off_track"].setChecked(True)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.ui_changed()


    def ui_changed(self, _=None):
        self.state_positions = {
                state_position_type: self._state_positions[state_position_type].isChecked()
                for state_position_type in self.state_position_types
        }

        self.filter_changed.emit()

    def filter(self, tp: TripPoint):
        if tp.state_position not in self.state_positions:
            print(tp.state_position)
        if self.state_positions.get(tp.state_position, True) == False:
            return False
        return True





class MainWind(QMainWindow):
    shown_data_changed_signal = pyqtSignal()
    showing_live_changed = pyqtSignal(bool)

    def __init__(self):
        QMainWindow.__init__(self)


        self.map = MapWidget()
        self.setCentralWidget(self.map)
        self._menu_data = self.menuBar().addMenu(self.tr("&Data"))
        self._menu_data.addAction("Set date/time", self.set_date_time)

        uri = "Point?field=vehicle_type:string(100)&field=route_short_name:string(20)&field=oldness:int&field=rotation:real&index=yes"
        self.data_layer = QgsVectorLayer(uri,
                           'PID realtime',
                           'memory')
        # d = QDomDocument()
        # d.setContent(open("labeling.xml").read())
        # labeling = QgsAbstractVectorLayerLabeling.create(d.documentElement(), QgsReadWriteContext())
        # self.data_layer.setLabeling(labeling)
        #self.data_layer.setLabelsEnabled(True)
        #print(self.data_layer.loadSldStyle("style.xml"))
        print(self.data_layer.loadNamedStyle("style.qml"))
        d = QDomDocument()
        d.setContent(open("labeling.xml").read())
        labeling = QgsAbstractVectorLayerLabeling.create(d.documentElement(), QgsReadWriteContext())
        self.data_layer.setLabeling(labeling)
        self.data_layer.setLabelsEnabled(True)
        self.data_layer.setLabelsEnabled(True)
        self.data_prov = self.data_layer.dataProvider()

        # set the map canvas layer set
        self.map.canvas.setLayers([self.data_layer, orm, mapy_cz])
        self.map.addLayer(self.data_layer)


        self.feature_identifier = MainTool(self, self.map.canvas)
        #self.feature_identifier.setLayer(self.data_layer)
        #self.feature_identifier.featureIdentified.connect(self.feature_clicked)
        self.map.canvas.setMapTool(self.feature_identifier)

        self.current_data = []
        self.current_filtered_data = []
        self.current_data_capture_time = None
        self.showing_live = False

        self._tmp = QPushButton()
        self._tmp.clicked.connect(self.tmp)
        self._tmp2 = QPushButton()
        self._tmp2.clicked.connect(self.tmp2)

        # dockWidget = QDockWidget("Dock Widget")
        # dockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        # dockWidget.setWidget(self._tmp)
        # self.addDockWidget(Qt.LeftDockWidgetArea, dockWidget)

        # dockWidget = QDockWidget("Dock Widget")
        # dockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        # dockWidget.setWidget(self._tmp2)
        # self.addDockWidget(Qt.LeftDockWidgetArea, dockWidget)

        self._time_controll = TimeControllToolBar(self)
        toolbar = self.addToolBar(self._time_controll)
        # dockWidget = QDockWidget("Time controll")
        # dockWidget.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum);
        # dockWidget.setAllowedAreas(Qt.TopDockWidgetArea | Qt.BottomDockWidgetArea)
        # dockWidget.setWidget(self._time_controll)
        # self.addDockWidget(Qt.TopDockWidgetArea, dockWidget)

        self._filter = FilterWidget(self)
        self._filter.filter_changed.connect(self.show_parsed_data)
        dockWidget = QDockWidget("Filter")
        dockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        dockWidget.setWidget(self._filter)
        self.addDockWidget(Qt.LeftDockWidgetArea, dockWidget)

        self._history_widgets = []

    @asyncSlot()
    async def tmp2(self):
        dt = datetime.datetime(2024, 7, 16, 8, 0, 0)
        g = gtfs.for_date(dt)
        await g.load_trips()
        #await g.load_stops_for_all_trips()
        trips = [Trip(t.id, dt.date()) for t in g.trips_by_routes["L22"] if t.direction == 1]
        ths = {t.trip_id: TripHistory(t) for t in trips}

        main_th = ths['22_28258_240701']

        c = await get_communication()
        #dts = await c.list_realtime_data(datetime.datetime.combine(dt.date(), datetime.time(8,0,0)), datetime.datetime.combine(dt.date(), datetime.time(9,0,0)))
        #for dt in dts:
        #    print("GET", dt)
        #    for fname, data in (await c.get_data(dt)).items():
        #        proc = await asyncio.create_subprocess_exec("gunzip", stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE)
        #        stdout, stderr = await proc.communicate(data)
        #        data = json.loads(stdout)
        #        for dato in data["features"]:
        #            trip_id = dato["properties"]["trip"]["gtfs"]["trip_id"]
        #            if trip_id in ths:
        #                await ths[trip_id].load_gtfs_shape()
        #                ths[trip_id].add_history_point(dt, dato)
        #        break

        data, source_timestamps = await c.get_preprocessed_data(dt, "22")
        print("P1")
        data = await unzip_parse(data)
        print("P3")

        for trip_id, th in ths.items():
            if trip_id in data:
                th.add_preprocessed_data(data[trip_id])

        print("P4")



        await main_th.load_stops(await c.gtfs_get_stop_times(datetime.datetime.combine(main_th.trip.date, datetime.time()), main_th.trip.trip_id)) # HACK: Date is not CBOR supported type
        #for th in ths.values():
            #await th.load_stops()

        print("P5")

        history_graph = graph.HistoryGraph(main_th.stops, "22", dt-datetime.timedelta(hours=2), dt)

        history_graph.add_trip(main_th)
        print("P7")
        history_graph.show()
        tmp_windows[history_graph]=history_graph

    @asyncSlot()
    async def tmp(self):
        dt = datetime.datetime(2024, 7, 4, 8, 4, 20)
        await self.load_data(dt)
        t = Trip('156_324_240701', dt.date())
        th = TripHistory(t)
        await th.load_gtfs_shape()
        await th.load_history(dt-datetime.timedelta(minutes=45), dt)
        self.plot_gtfs_shape(th.trip, th.gtfs_shape)
        self.plot_history_shape_mapping(th)
        self.plot_history(th)

        history_graph = HistoryGraph(th.stops)
        history_graph.add_trip(th)
        history_graph.show()
        tmp_windows[history_graph]=history_graph


    def set_date_time(self):
        self.window_set_date_time = SetDateTimeWind(self)
        self.window_set_date_time.show()

    def tool_identifier(self):
        self.map.canvas.setMapTool(feature_identifier)

    def feature_clicked(self, feature):
        print(f"You clicked on feature {feature}")

    @asyncSlot()
    async def next_data(self):
        await self.set_live(False)
        c = await get_communication()
        r = await c.get_next_data(self.current_data_capture_time)
        if r is None:
            error("Couldn't move to next data, you are on the last data.")
        else:
            data = await unzip_parse(r[1])
            await gtfs.for_date(r[0]).load_stops()
            self.show_data(data, r[0])

    @asyncSlot()
    async def prev_data(self):
        await self.set_live(False)
        c = await get_communication()
        r = await c.get_prev_data(self.current_data_capture_time)
        if r is None:
            error("Couldn't move to previous data, you are on the first data.")
        else:
            data = await unzip_parse(r[1])
            await gtfs.for_date(r[0]).load_stops()
            self.show_data(data, r[0])

    @asyncSlot()
    async def toogle_live(self):
        await self.set_live(not self.showing_live)

    @asyncSlot()
    async def set_live(self, val):
        if self.showing_live != val:
            self.showing_live = val
            if self.showing_live:
                await self.start_live()
            else:
                await self.stop_live()
            self.showing_live_changed.emit(self.showing_live)

    async def start_live(self):
        if live_data is not None:
            await gtfs.for_date(live_dt).load_stops()
            self.show_data(live_data, live_dt)
        live_subscribers.add(self.live_data_arrived)
        spawn_or_remove_fetch_live_task()

    async def stop_live(self):
        live_subscribers.remove(self.live_data_arrived)
        spawn_or_remove_fetch_live_task()



    async def live_data_arrived(self, dt, data, _):
        if self.showing_live:
            await gtfs.for_date(dt).load_stops()
            self.show_data(data, dt)
        else:
            pprint("live data arrived but we are not in live mode")


    async def load_data_near(self, dt):
        c = await get_communication()
        dts = await c.list_realtime_data(dt-datetime.timedelta(seconds=20), dt+datetime.timedelta(seconds=20))
        if not len(dts):
            error("No data near {format_dt_to_user(dt)}")
        else:
            near_dt = min(dts, key=lambda x: abs((x - dt).total_seconds()))
            await self.load_data(near_dt)

    async def load_data(self, dt):
        await self.set_live(False)
        c = await get_communication()
        data = await unzip_parse(await c.get_data(dt))
        await gtfs.for_date(dt).load_stops()
        self.show_data(data, dt)

    def show_data(self, data, capture_time):
        current_data = []


        for i, dato in enumerate(data["features"]):
            coord = dato["geometry"]["coordinates"]
            point = QgsPointXY(coord[0], coord[1])
            tp = TripPoint(json=dato, capture_time=capture_time)
            tp.point = point
            current_data.append(tp)

        self.current_data = current_data
        self.current_data_capture_time = capture_time

        self.show_parsed_data()

    def show_parsed_data(self):
        g = gtfs.for_date_cache.get(self.current_data_capture_time.date(), None)
        stops = g.stops or {} if g else {}
        current_filtered_data = [tp for tp in self.current_data if self._filter.filter(tp)]

        feats = []

        self.current_filtered_data = current_filtered_data

        layer_fields = self.data_layer.fields()
        for tp in current_filtered_data:
            feat = QgsFeature(layer_fields)
            geometry = QgsGeometry.fromPointXY(tp.point)
            feat.setGeometry(geometry)
            dato_route_short_name = tp.json["properties"]["trip"]["gtfs"]["route_short_name"]
            feat["route_short_name"] = str(dato_route_short_name)
            dato_vehicle_type = tp.json["properties"]["trip"]["vehicle_type"]
            if dato_vehicle_type is None:
                if len(dato_route_short_name) and dato_route_short_name[0] in "RSLUV":
                    feat["vehicle_type"] = "vlak"
                else:
                    feat["vehicle_type"] = "None"
            else:
                feat["vehicle_type"] = str(dato_vehicle_type["description_cs"])
            feat["oldness"] = (self.current_data_capture_time - tp.origin_timestamp).total_seconds()
            last_stop = stops.get(tp.json["properties"]["last_position"]["last_stop"]["id"], None)
            next_stop = stops.get(tp.json["properties"]["last_position"]["next_stop"]["id"], None)
            if last_stop and next_stop:
                lat = next_stop.lat - last_stop.lat
                lon = next_stop.lon - last_stop.lon
                atan = math.atan2(lat, lon*data_utils.lon_muntiplicator)/math.pi*180
                feat["rotation"] = (360 - atan) % 360
            feats.append(feat)

        self.data_prov.truncate()
        self.data_prov.addFeatures(feats)
        self.data_prov.reloadData()
        print("reloadData done")
        self.data_layer.triggerRepaint()
        print("repaintRequested done")

        self.shown_data_changed_signal.emit()

    def add_gtfs_shape_layer(self, trip):
        uri = "LineString"
        shape_line_layer = QgsVectorLayer(uri,
                           f"Shape of {trip.trip_id}",
                           'memory')
        shape_line_layer.renderer().symbol().setWidth(1.3)
        shape_line_layer.renderer().symbol().setColor(QColor.fromRgb(0, 255, 0))

        self.map.addLayer(shape_line_layer)

        return shape_line_layer

    def add_stops_layer(self):
        uri = "Point?field=name:string(100)&index=yes"
        layer = QgsVectorLayer(uri,
                           f"Stops",
                           'memory')
        self.map.addLayer(layer)
        return layer

    def add_history_layers(self,  th: TripHistory):
        uri = "LineString"
        history_line_layer = QgsVectorLayer(uri,
                           f"History of {th.trip.trip_id}",
                           'memory')
        history_line_layer.renderer().symbol().setWidth(1.3)
        history_line_layer.renderer().symbol().setColor(QColor.fromRgb(255, 0, 0))

        uri = "Point?field=repeated:int&field=without_data:int&index=yes"
        history_points_layer = QgsVectorLayer(uri,
                           f"History of {th.trip.trip_id}",
                           'memory')
        d = QDomDocument()
        d.setContent(open("labeling-history-points.xml").read())
        labeling = QgsAbstractVectorLayerLabeling.create(d.documentElement(), QgsReadWriteContext())
        history_points_layer.setLabeling(labeling)
        history_points_layer.setLabelsEnabled(True)

        self.map.addLayer(history_line_layer)
        self.map.addLayer(history_points_layer)

        return history_line_layer, history_points_layer



    def add_history_shape_mapping_layer(self, th: TripHistory):
        uri = "LineString"
        layer = QgsVectorLayer(uri,
                           f"Shape history mapping of {th.trip.trip_id}",
                           'memory')
        layer.renderer().symbol().setWidth(1.3)
        layer.renderer().symbol().setColor(QColor.fromRgb(0, 0, 255))
        self.map.addLayer(layer)
        return layer

    def add_trip_history_widget(self, trip):
        widget = TripHistoryWidget(trip, self)
        dockWidget = DockWidgetWithClosedSignal("History of trip")
        dockWidget.closed.connect(widget.remove)
        dockWidget.closed.connect(lambda: self._history_widgets.remove((dockWidget, widget)))
        dockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        dockWidget.setWidget(widget)
        self.addDockWidget(Qt.LeftDockWidgetArea, dockWidget)
        self._history_widgets.append((dockWidget, widget))
        return widget




class SetDateTimeCaptureLabel(ClickableLabel):
    def __init__(self, dt, window_set_date_time):
        QLabel.__init__(self, str(dt.second))
        self.window_set_date_time = window_set_date_time
        self.dt = dt

        self.setStyleSheet("color: green;")
        self.clicked.connect(self.on_click)


    @asyncSlot()
    async def on_click(self):
        print(self.dt)
        self.window_set_date_time.hide()
        await self.window_set_date_time.window_main.load_data(self.dt)
        self.window_set_date_time.window_main = None

class SetDateTimeHour(ClickableLabel):
    def __init__(self, hour, window_set_date_time):
        QLabel.__init__(self, str(hour))
        self.window_set_date_time = window_set_date_time
        self.hour = hour

        self.clicked.connect(self.on_click)
        self.restyle()

    def restyle(self):
        self.setStyleSheet("font-weight: bold;" if self.hour == self.window_set_date_time.hour else "")
        print(self.hour, self.window_set_date_time.hour)

    def on_click(self):
        self.window_set_date_time.hour = self.hour
        self.window_set_date_time.master_changed()


def clearlayout(layout):
    for i in reversed(range(layout.count())):
        layout.removeItem(layout.itemAt(i))


class SetDateTimeWind(QWidget):
    def __init__(self, window_main):
        QWidget.__init__(self)
        self.window_main = window_main

        self.hour = 0

        self._calendar = QCalendarWidget()
        self._calendar.setFirstDayOfWeek(Qt.Monday)
        self._calendar.selectionChanged.connect(self.master_changed)

        self._hour_layout = QGridLayout()
        self._hours = [SetDateTimeHour(i, self) for i in range(24)]
        for i, x in enumerate(self._hours):
            self._hour_layout.addWidget(x, i//6, i%6)


        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._layout_calendar = QHBoxLayout()
        self._layout.addWidget(self._calendar)
        self._layout.addLayout(self._layout_calendar)
        self._layout_calendar.addWidget(self._calendar)
        self._layout_calendar.addLayout(self._hour_layout)

        self._scroll = QScrollArea()
        self._scroll.setWidgetResizable(True)
        self._layout.addWidget(self._scroll)
        self._scroll_widget = QWidget()
        self._scroll.setWidget(self._scroll_widget)
        self._scroll_layout = QGridLayout()
        self._scroll_widget.setLayout(self._scroll_layout)
        # self._layout.addLayout(self._scroll_layout)

        self._minutes_labels = [QLabel(str(i)) for i in range(60)]
        for x in self._minutes_labels:
            x.setStyleSheet("font-weight: bold;")
        for i, x in enumerate(self._minutes_labels):
            self._scroll_layout.addWidget(x, i, 0)

        self._10s_layouts = [[QHBoxLayout() for _ in range(6)] for _ in range(60)]
        for i, x in enumerate(self._10s_layouts):
            for j, y in enumerate(x):
                self._scroll_layout.addLayout(y, i, j+1)

        self._10s_labels = [[[] for j in range(6)] for _ in range(60)]
        self.master_changed()

    def update_10s_layouts(self, _10s_labels):

        for xa, xb in zip(self._10s_layouts, self._10s_labels):
            for ya, yb in zip(xa, xb):
                for z in yb:
                    z.clear()
                clearlayout(ya)

        self._10s_labels = _10s_labels

        for xa, xb in zip(self._10s_layouts, self._10s_labels):
            for ya, yb in zip(xa, xb):
                for z in yb:
                    ya.addWidget(z)

    @asyncSlot()
    async def master_changed(self):
        for x in self._hours:
            x.restyle()
        date = datetime.datetime.combine(self._calendar.selectedDate().toPyDate(), datetime.time(self.hour,0,0))

        self.update_10s_layouts([[[QLabel(f"{j}0")] for j in range(6)] for _ in range(60)])

        c = await get_communication()

        _10s_labels = [[[] for j in range(6)] for _ in range(60)]
        dts = await c.list_realtime_data(date, date+datetime.timedelta(hours=1))

        if date != datetime.datetime.combine(self._calendar.selectedDate().toPyDate(), datetime.time(self.hour,0,0)):
            return # Something changed during await

        for dt in dts:
            _10s_labels[dt.minute][dt.second//10].append(SetDateTimeCaptureLabel(dt, self))

        self.update_10s_layouts(_10s_labels)




class SmallTripPointLabel(ClickableLabel):
    def __init__(self, tp: TripPoint, on_click=None):
        super().__init__()
        if on_click:
            self.on_click = on_click
            self.clicked.connect(self.on_click)
        self.tp = tp
        route_short_name = tp.json["properties"]["trip"]["gtfs"]["route_short_name"]
        self.setText(f"{route_short_name} – {tp.trip.trip_id} ({tp.state_position})")

class SmallTripPointLabelSelector(SmallTripPointLabel):
    def __init__(self, tp: TripPoint, trip_point_selector):
        super().__init__(tp)
        self.trip_point_selector = trip_point_selector
        self.clicked.connect(self.on_click)

    def on_click(self):
        TripPointWindow(self.tp, self.trip_point_selector.mwnd)
        self.trip_point_selector.hide()



class TripPointSelector(QWidget):
    def __init__(self, tps, mwnd):
        self.mwnd = mwnd
        super().__init__()
        self.tps = tps
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        self._labels = [SmallTripPointLabelSelector(tp, self) for tp in tps]
        for l in self._labels:
            self._layout.addWidget(l)
        self.setWindowTitle(f"{window_title_prefix} - point selector")
        self.show()
        tmp_windows[self] = self


tmp_windows = {}

class BoldLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        f = self.font()
        f.setBold(True)
        self.setFont(f)

class TimeLabel(ClickableLabel):
    def __init__(self, mwnd, dt=None):
        super().__init__()
        self.set_dt(dt)
        self.clicked.connect(self.on_click)
        self.mwnd = mwnd

    def set_dt(self, dt):
        if dt:
            self.setText(format_dt_to_user(dt))
        self.dt = dt


    @asyncSlot()
    async def on_click(self):
        if self.dt:
            await self.mwnd.load_data_near(self.dt)



class TripPointWindow(QWidget):
    def __init__(self, tp, mwnd):
        super().__init__()
        self.mwnd = mwnd
        self._layout = QVBoxLayout()

        self._info_layout = QGridLayout()
        self._layout.addLayout(self._info_layout)
        i = 0

        def add_info_widgets(title, val=None):
            nonlocal i
            self._info_layout.addWidget(title, i, 0)
            if val:
                self._info_layout.addWidget(val, i, 1)
            i += 1

        self._info_capture_time = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("Capture time"), self._info_capture_time)
        self._info_first_capture_time = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("First capture time"), self._info_first_capture_time)
        self._info_last_capture_time = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("Last capture time"), self._info_last_capture_time)
        self._info_origin_time = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("Origin time"), self._info_origin_time)

        add_info_widgets(BoldLabel("Trip"))
        self._info_headsig = QLabel()
        add_info_widgets(QLabel("Headsign"), self._info_headsig)
        self._info_trip_id = QLabel()
        add_info_widgets(QLabel("Trip ID"), self._info_trip_id)
        self._info_trip_name = QLabel()
        add_info_widgets(QLabel("Trip name"), self._info_trip_name)
        self._info_route = QLabel()
        add_info_widgets(QLabel("Route"), self._info_route)
        self._info_agency = QLabel()
        add_info_widgets(QLabel("Agency"), self._info_agency)
        self._info_sequence = QLabel()
        add_info_widgets(QLabel("Sequence"), self._info_sequence)
        self._info_vehicle_type = QLabel()
        add_info_widgets(QLabel("Vehicle"), self._info_vehicle_type)
        self._info_start_timestamp = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("Start timestamp"), self._info_start_timestamp)
        self._info_start_timestamp = TimeLabel(self.mwnd)
        add_info_widgets(QLabel("Start "), self._info_start_timestamp)

        def add_stop_info():
            class StopInfo:
                ...
            obj = StopInfo()
            obj.name = QLabel()
            add_info_widgets(QLabel("Name"), obj.name)
            obj.id = QLabel()
            add_info_widgets(QLabel("Id"), obj.id)
            obj.arrival = TimeLabel(self.mwnd)
            add_info_widgets(QLabel("Arrival"), obj.arrival)
            obj.departure = TimeLabel(self.mwnd)
            add_info_widgets(QLabel("Departure"), obj.departure)
            return obj


        add_info_widgets(BoldLabel("OpenAPI Last stop"))
        self._info_last_stop = add_stop_info()
        add_info_widgets(BoldLabel("OPpenAPI Next stop"))
        self._info_next_stop = add_stop_info()

        add_info_widgets(BoldLabel("OpenAPI delay"))
        self._info_openapi_delay_arrival = QLabel()
        add_info_widgets(QLabel("Last stop arrival"), self._info_openapi_delay_arrival)
        self._info_openapi_delay_departure = QLabel()
        add_info_widgets(QLabel("Last stop departure"), self._info_openapi_delay_departure)
        self._info_openapi_delay = QLabel()
        add_info_widgets(QLabel("Now"), self._info_openapi_delay)

        add_info_widgets(BoldLabel("Shape & position"))
        self._info_state = QLabel()
        add_info_widgets(QLabel("State"), self._info_state)
        self._info_shape_dist = QLabel()
        add_info_widgets(QLabel("Shape dist"), self._info_shape_dist)
        self._info_openapi_shape_dist = QLabel()
        add_info_widgets(QLabel("OpenAPI shape dist"), self._info_openapi_shape_dist)



        #self._info_rotation = QLabel()
        #add_info_widgets(QLabel("Rotation"), self._info_rotation)

        del i

        self.setLayout(self._layout)
        self.tp = tp
        self.setLayout(self._layout)
        self.setWindowTitle(f"{window_title_prefix} - trip point")
        self.show()
        tmp_windows[self] = self

        self._plot_history_button = QPushButton("Plot history")
        self._layout.addWidget(self._plot_history_button)
        self._plot_history_button.clicked.connect(self.plot_history)

        self._show_json = QPushButton("Show json")
        self._layout.addWidget(self._show_json)
        self._show_json.clicked.connect(self.show_json)


        self.fill_data()

    def show_json(self):
        w = QTextEdit()
        w.setAcceptRichText(False)
        w.setText(pprint.pformat(self.tp.json))
        w.resize(600,700);
        w.setWindowTitle(f"{window_title_prefix} - raw json")
        tmp_windows[w] = w
        w.show()


    @asyncSlot()
    async def fill_data(self):
        tp = self.tp

        self._info_origin_time.set_dt(tp.origin_timestamp)
        self._info_capture_time.set_dt(tp.capture_time)
        #self._info_origin_time.setText(self._info_origin_time.text() + (tp.capture_time - tp.origin_timestamp).total_seconds()} s before capture")

        self._info_openapi_delay.setText(format_delay_s(tp.openapi_delay["actual"]))
        self._info_openapi_delay_arrival.setText(format_delay_s(tp.openapi_delay["last_stop_arrival"]))
        self._info_openapi_delay_departure.setText(format_delay_s(tp.openapi_delay["last_stop_departure"]))

        self._info_start_timestamp.set_dt(tp.start_timestamp)

        trip_json = tp.json["properties"]["trip"]
        self._info_trip_id.setText(tp.trip.trip_id)
        self._info_trip_name.setText(trip_json["gtfs"]["trip_short_name"] or "")
        self._info_headsig.setText(trip_json["gtfs"]["trip_headsign"])
        self._info_route.setText(trip_json["origin_route_name"] or trip_json["gtfs"]["route_short_name"])
        self._info_vehicle_type.setText((trip_json["vehicle_type"]["description_en"] if trip_json["vehicle_type"] else "undef")
                + (" ❄️" if trip_json["air_conditioned"] else "")
                + (" 🔌" if trip_json["usb_chargers"] else "")
                + (" ♿" if trip_json["wheelchair_accessible"] else "")
                                        )
        self._info_sequence.setText(str(trip_json["sequence_id"]))

        self._info_openapi_shape_dist.setText(str(tp.openapi_shape_dist_traveled))

        self._info_agency.setText(str(trip_json['agency_name']['scheduled']) + " / " + str(trip_json['agency_name']['real']))

        self._info_state.setText(tp.state_position)





        g = gtfs.for_date(tp.origin_timestamp)
        await g.load_stops()

        def fill_stop(json, obj):
            if json is not None:
                stop = g.stops.get(json["id"], None)
                obj.id.setText(json["id"])
                obj.name.setText(stop.name if stop else "")
                obj.arrival.set_dt(datetime.datetime.fromisoformat(json["arrival_time"]) if json["arrival_time"] else None)
                obj.departure.set_dt(datetime.datetime.fromisoformat(json["departure_time"]) if json["departure_time"] else None)
            else:
                obj.name.setText("")
                obj.arrival.set_dt(None)
                obj.departure.set_dt(None)

        fill_stop(tp.json["properties"]["last_position"]["last_stop"], self._info_last_stop)
        fill_stop(tp.json["properties"]["last_position"]["next_stop"], self._info_next_stop)

        #stops = g.stops
        #last_stop = stops.get(tp.json["properties"]["last_position"]["last_stop"]["id"], None)
        #next_stop = stops.get(tp.json["properties"]["last_position"]["next_stop"]["id"], None)
        #if last_stop and next_stop:
            #lat = next_stop.lat - last_stop.lat
            #lon = next_stop.lon - last_stop.lon
            #atan = math.atan2(lat, lon*data_utils.lon_muntiplicator)/math.pi*180
            #rotation = - atan
            #self._info_rotation.setText(f"{lat} {lon} -> {atan} -> {rotation}")

    @asyncSlot()
    async def plot_history(self):
        history_widget = self.mwnd.add_trip_history_widget(self.tp.trip)
        await history_widget.load_gtfs_shape()
        await history_widget.load_stops()
        await history_widget.load(self.tp.capture_time-datetime.timedelta(hours=0.1), self.tp.capture_time)

        #await self.mwnd.plot_gtfs_shape(self.tp)
        #await self.mwnd.plot_history(self.tp, self.tp.capture_time-datetime.timedelta(minutes=30), self.tp.capture_time)

class LayerWidget(QWidget):
    def __init__(self, layers, title, mwnd):
        self.mwnd = mwnd
        super().__init__()
        self.layers = layers

        self._layout = no_space(QHBoxLayout())
        self.setLayout(self._layout)

        self._visible = QCheckBox()
        self._visible.setCheckState(Qt.Checked)
        self._visible.stateChanged.connect(self.set_visible)
        self._layout.addWidget(self._visible)

        self._title = QLabel(title)
        self._layout.addWidget(self._title)

    def set_visible(self, val):
        for l in self.layers:
            if val:
                self.mwnd.map.addLayer(l)
            else:
                self.mwnd.map.removeLayer(l)


class SetRangeWindow(QWidget):
    def __init__(self, dt_from, dt_to, callback):
        super().__init__()
        self.callback = callback
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._dt_from = QLineEdit(format_dt_to_user(dt_from))
        self._layout.addWidget(self._dt_from)
        self._dt_from.returnPressed.connect(self.done)

        self._dt_to = QLineEdit(format_dt_to_user(dt_to))
        self._layout.addWidget(self._dt_to)
        self._dt_to.returnPressed.connect(self.done)

        self._done = QPushButton("(Re)load")
        self._done.clicked.connect(self.done)
        self._layout.addWidget(self._done)

        self.setWindowTitle(f"{window_title_prefix} - range selector")

    @asyncSlot()
    async def done(self):
        dt_from = parse_dt_from_user(self._dt_from.text())
        dt_to = parse_dt_from_user(self._dt_to.text())
        self.hide()
        del tmp_windows[self]
        await self.callback(dt_from, dt_to)




class LoadRangeWidget(QWidget):
    def __init__(self, mwnd, route_id):
        super().__init__()
        self.route_id = route_id
        self.cb_clear = lambda: None
        self.cb_add_preprocessed_data = lambda data, dt_from, dt_to: None
        self.cb_data_changed = lambda: None
        self.cb_add_live_data = lambda: error("Live not supported")

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        self._times_layout = QGridLayout()
        self._layout.addLayout(self._times_layout)

        self.dt_from = self.dt_to = None

        self.live = False
        self._dt_from = TimeLabel(mwnd)
        self._dt_to = TimeLabel(mwnd)
        self._times_layout.addWidget(QLabel("Loaded from:"), 0, 0)
        self._times_layout.addWidget(self._dt_from, 0, 1)
        self._times_layout.addWidget(QLabel("Loaded to:"), 1, 0)
        self._times_layout.addWidget(self._dt_to, 1, 1)

        self._buttons = QToolBar()
        self._layout.addWidget(self._buttons)
        self._load_previous = QAction(QIcon(":images/themes/default/temporal_navigation/rewindToStart.svg"), "Load previous")
        self._load_previous.triggered.connect(self.load_prev)
        self._buttons.addAction(self._load_previous)
        self._reload = QAction(QIcon(":images/themes/default/mActionReload.svg"), "Reload all")
        self._buttons.addAction(self._reload)
        self._reload.triggered.connect(self.select_and_reload)
        self._load_next = QAction(QIcon(":images/themes/default/temporal_navigation/skipToEnd.svg"), "Load next")
        self._buttons.addAction(self._load_next)
        self._load_next.triggered.connect(self.load_next)
        self._start_live = QAction(QIcon(":images/themes/default/temporal_navigation/forward.svg"), "Load to now and start live updates")
        self._buttons.addAction(self._start_live)
        self._start_live.triggered.connect(self.start_live)
        self._stop_live = QAction(QIcon(":images/themes/default/temporal_navigation/pause.svg"), "Stop live updates")
        self._buttons.addAction(self._stop_live)
        self._stop_live.triggered.connect(self.stop_live)

    def update_dt_labels(self):
        self._dt_from.set_dt(self.dt_from)
        self._dt_to.set_dt(self.dt_to)

    async def clear(self):
        self.dt_from = self.dt_to = None
        self.update_dt_labels()
        await self.cb_clear()

    @asyncSlot()
    async def load_prev(self):
        await self.load(self.dt_from-datetime.timedelta(minutes=15), self.dt_from-datetime.timedelta(seconds=20))

    @asyncSlot()
    async def load_next(self):
        await self.load(self.dt_to+datetime.timedelta(seconds=10), self.dt_to+datetime.timedelta(minutes=15))

    def select_and_reload(self):
        async def callback(dt_from, dt_to):
            self.stop_live()
            await self.clear()
            await self.load(dt_from, dt_to, strict=True)

        w = SetRangeWindow(self.dt_from, self.dt_to+datetime.timedelta(seconds=10), callback)
        tmp_windows[w] = w
        w.show()

    @asyncSlot()
    async def start_live(self):
        if not self.live:
            if (datetime.datetime.now(local_timezone) - self.dt_to).total_seconds() > 60*60:
                error("Last data is too old to start live")
                return
            await self.load(self.dt_to, datetime.datetime.now(local_timezone), send_changed=False)
            if (datetime.datetime.now(local_timezone) - self.dt_to).total_seconds() > 30:
                error("Couldn't fetch data just before now, couldn't continue with live.")
                return
            self.live = True
            live_subscribers.add(self.live_data_arrived)
            spawn_or_remove_fetch_live_task()
            if live_data:
                await self.cb_add_live_data(live_dt, live_data, live_data_by_trip_id)
            await self.cb_data_changed()

    def stop_live(self):
        if self.live:
            self.live = False
            live_subscribers.remove(self.live_data_arrived)
            spawn_or_remove_fetch_live_task()

    async def live_data_arrived(self, dt, data, data_by_trip_id):
        if self.live:
            await self.cb_add_live_data(dt, data, data_by_trip_id)
            self.dt_to = dt
            self.update_dt_labels()
        else:
            pprint("live data arrived but we are not in live mode")

    async def load(self, dt_from, dt_to, strict=False, send_changed=True):
        c = await get_communication()
        dt = dt_from.replace(second=0, microsecond=0, minute=0)
        while dt < dt_to:
            data, source_timestamps = await c.get_preprocessed_data(dt, self.route_id)
            if strict:
                source_timestamps = [x for x in source_timestamps if dt_from <= x and x < dt_to]
            data = await unzip_parse(data)
            await self.cb_add_preprocessed_data(data, dt_from if strict else dt_minf, dt_to if strict else dt_pinf)
            if len(source_timestamps):
                print(self.dt_from, self.dt_to)
                print(source_timestamps[0], source_timestamps[-1])
                self.dt_from = min(source_timestamps[0], self.dt_from) if self.dt_from else source_timestamps[0]
                self.dt_to = max(source_timestamps[-1], self.dt_to) if self.dt_to else source_timestamps[-1]
                print(self.dt_from, self.dt_to)
                self.update_dt_labels()
            dt += datetime.timedelta(hours=1)
        if send_changed:
            await self.cb_data_changed()




class TripHistoryWidget(QWidget):
    def __init__(self, trip, mwnd):
        super().__init__()
        self.mwnd = mwnd
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._title = QLabel(trip.trip_id)
        f = self._title.font()
        f.setPixelSize(18)
        self._title.setFont(f)
        self._layout.addWidget(self._title)

        self.th = TripHistory(trip)
        self.trip = trip

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.gtfs_shape_layer = mwnd.add_gtfs_shape_layer(self.trip)
        self.stops_layer = mwnd.add_stops_layer()
        self.history_shape_mapping_layer = mwnd.add_history_shape_mapping_layer(self.th)
        self.history_line_layer, self.history_points_layer = mwnd.add_history_layers(self.th)

        self._history_layers = LayerWidget([self.history_line_layer, self.history_points_layer], "History layer", mwnd)
        self._layout.addWidget(self._history_layers)

        self._history_shape_mapping_layer = LayerWidget([self.history_shape_mapping_layer], "History-shape mapping", mwnd)
        self._layout.addWidget(self._history_shape_mapping_layer)

        self._gtfs_shape_layer = LayerWidget([self.gtfs_shape_layer], "Shape layer", mwnd)
        self._layout.addWidget(self._gtfs_shape_layer)

        self._stops_layer = LayerWidget([self.stops_layer], "Stops layer", mwnd)
        self._layout.addWidget(self._stops_layer)

        self._load_range = LoadRangeWidget(self.mwnd, self.trip.trip_id.split("_")[0])
        self._load_range.cb_add_preprocessed_data = self.add_preprocessed_data
        self._load_range.cb_clear = self.clear_history
        self._load_range.cb_data_changed = self.redraw_all_async
        self._load_range.cb_add_live_data = self.add_live_data
        self._layout.addWidget(self._load_range)


        self._show_graph = QPushButton("Show graph")
        self._show_graph.clicked.connect(self.show_graph)
        self._layout.addWidget(self._show_graph)

        self.graph = None

    def remove(self):
        self._load_range.stop_live()
        self.mwnd.map.removeLayer(self.gtfs_shape_layer)
        self.mwnd.map.removeLayer(self.stops_layer)
        self.mwnd.map.removeLayer(self.history_shape_mapping_layer)
        self.mwnd.map.removeLayer(self.history_line_layer)
        self.mwnd.map.removeLayer(self.history_points_layer)


    async def clear_history(self):
        self.th.history = []

    async def add_preprocessed_data(self, data, dt_from, dt_to):
        if self.trip.trip_id in data:
            self.th.add_preprocessed_data(data[self.trip.trip_id], dt_from, dt_to)

    async def add_live_data(self, dt, data, data_by_trip_id):
        if self.trip.trip_id in data_by_trip_id:
            self.th.add_history_point(dt, data_by_trip_id[self.trip.trip_id])
            self.redraw_all()


    async def load(self, dt_from, dt_to):
        await self._load_range.load(dt_from, dt_to)
        return
        c = await get_communication()
        dt = dt_from.replace(second=0, microsecond=0, minute=0)
        while dt < dt_to:
            data, source_timestamps = await c.get_preprocessed_data(dt, self.trip.trip_id.split("_")[0])
            print("P1")
            data = await unzip_parse(data)
            print("P3")
            dt += datetime.timedelta(hours=1)


    async def load_stops(self):
        print("LOAD STOPS")
        c = await get_communication()
        print(self.trip.date)
        await self.th.load_stops(await c.gtfs_get_stop_times(datetime.datetime.combine(self.trip.date, datetime.time()), self.trip.trip_id)) # HACK: Date is not CBOR supported type
        self.redraw_stops_layer()

    def redraw_history_layers(self):
        th = self.th
        history_line_prov = self.history_line_layer.dataProvider()
        history_points_prov = self.history_points_layer.dataProvider()

        history_line_fields = self.history_line_layer.fields()
        history_points_fields = self.history_points_layer.fields()

        feat = QgsFeature(history_line_fields)
        history_points_feats = []
        line_points = []
        for hp in th.history:
            point = QgsPointXY(hp.lon, hp.lat)
            line_points.append(point)
        geometry = QgsGeometry.fromPolylineXY(line_points)
        feat.setGeometry(geometry)
        history_line_prov.truncate()
        history_line_prov.addFeatures([feat])


        for hp in th.history:
            feat = QgsFeature(history_points_fields)
            point = QgsPointXY(hp.lon, hp.lat)
            geometry = QgsGeometry.fromPointXY(point)
            feat.setGeometry(geometry)
            # feat["repeated"] = tp_data["repeated"]
            # feat["without_data"] = tp_data["without_data"]
            history_points_feats.append(feat)
        history_points_prov.truncate()
        history_points_prov.addFeatures(history_points_feats)

        layer_updated(self.history_line_layer)
        layer_updated(self.history_points_layer)

    def redraw_history_shape_mapping_layer(self):
        th = self.th

        layer = self.history_shape_mapping_layer
        prov = layer.dataProvider()

        fields = layer.fields()
        feats = []
        for hp in th.history:
            if hp.shape_point is not None:
                feat = QgsFeature(fields)
                x = shape_indexer(th.gtfs_shape, hp.shape_point)
                points = [
                        QgsPointXY(hp.lon, hp.lat),
                        QgsPointXY(x[1], x[0])
                ]
                geometry = QgsGeometry.fromPolylineXY(points)
                feat.setGeometry(geometry)
                feats.append(feat)

        prov.truncate()
        prov.addFeatures(feats)

        layer_updated(layer)

    def redraw_gtfs_shape_layer(self):
        shape = self.th.gtfs_shape
        shape_line_layer = self.gtfs_shape_layer

        shape_line_prov = shape_line_layer.dataProvider()

        shape_line_fields = shape_line_layer.fields()
        feat = QgsFeature(shape_line_fields)
        points = []
        for x in shape:
            points.append(QgsPointXY(x[1], x[0]))
        geometry = QgsGeometry.fromPolylineXY(points)
        feat.setGeometry(geometry)
        shape_line_prov.truncate()
        shape_line_prov.addFeatures([feat])

        shape_line_layer.updateExtents()
        shape_line_layer.triggerRepaint()
        shape_line_prov.reloadData()


        shape_line_layer.triggerRepaint()

    def redraw_stops_layer(self):
        layer = self.stops_layer

        prov = layer.dataProvider()
        fields = layer.fields()

        feats = []
        for stop in self.th.stops:
            feat = QgsFeature(fields)
            point = QgsPointXY(stop.stop.lon, stop.stop.lat)
            geometry = QgsGeometry.fromPointXY(point)
            feat.setGeometry(geometry)
            feat["name"] = stop.stop.name
            feats.append(feat)

        prov.truncate()
        prov.addFeatures(feats)

        layer_updated(self.history_line_layer)


    def redraw_all(self):
        self.redraw_history_layers()
        self.redraw_history_shape_mapping_layer()

    async def redraw_all_async(self):
        self.redraw_all()

    async def load_gtfs_shape(self):
        await self.th.load_gtfs_shape()
        self.redraw_gtfs_shape_layer()

    def clear_graph(self): self.graph = None
    def show_graph(self):
        if self.graph is None:
            print(self.th.stops)
            self.graph = graph.HistoryGraph(self.th.stops, self.th.trip.trip_id.split("_")[0])
            self.graph.closed.connect(self.clear_graph)
            self.graph.add_trip(self.th)
            self.graph.show()
        else:
            self.graph.activateWindow()





class MainTool(QgsMapTool):
    def __init__(self, wnd, canvas):
        QgsMapTool.__init__(self, canvas)
        self.wnd = wnd

    def canvasClicked(self, e):
        pt = e.mapPoint()
        data_by_dist = sorted([(crs_transform.transform(x.point).distance(pt), x) for x in self.wnd.current_filtered_data], key=lambda x:x[0])[:10]

        scale = self.wnd.map.canvas.scale()
        data_near = [(d, x) for d, x in data_by_dist if d/scale < 0.005]
        if len(data_near) == 1:
            tp_selector = TripPointWindow(data_near[0][1], self.wnd)
        if len(data_near) > 1:
            tp_selector = TripPointSelector([x for d, x in data_near], self.wnd)
        for d, x in data_near:
            print()
            print(d/scale, pt, x.point)
            pprint.pp(x.json)
            print()


    #def canvasMoveEvent(self, e):
        #print("canvasMoveEvent", e)

    def canvasPressEvent(self, e):
        print("canvasPressEvent", e)
        self.canvasClicked(e)

    def canvasReleaseEvent(self, e):
        print("canvasReleaseEvent", e)





class MapWidget(QWidget):
    def __init__(self):
        QWidget.__init__(self)

        self.canvas = QgsMapCanvas()
        self.canvas.setCanvasColor(Qt.white)
        self.canvas.enableAntiAliasing(True)

        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)
        self._layout.addWidget(self.canvas)

        self.actionZoomIn = QAction("Zoom in", self)
        self.actionZoomOut = QAction("Zoom out", self)
        self.actionPan = QAction("Pan", self)

        self.actionZoomIn.setCheckable(True)
        self.actionZoomOut.setCheckable(True)
        self.actionPan.setCheckable(True)

        self.actionZoomIn.triggered.connect(self.zoomIn)
        self.actionZoomOut.triggered.connect(self.zoomOut)
        #self.canvas.xyCoordinates.connect(self.xyCoordinates)
        self.actionPan.triggered.connect(self.pan)


        #self.toolbar = self.addToolBar("Canvas actions")
        #self.toolbar.addAction(self.actionZoomIn)
        #self.toolbar.addAction(self.actionZoomOut)
        #self.toolbar.addAction(self.actionPan)

        # create the map tools
        self.toolPan = QgsMapToolPan(self.canvas)
        self.toolPan.setAction(self.actionPan)
        self.toolZoomIn = QgsMapToolZoom(self.canvas, False) # false = in
        self.toolZoomIn.setAction(self.actionZoomIn)
        self.toolZoomOut = QgsMapToolZoom(self.canvas, True) # true = out
        self.toolZoomOut.setAction(self.actionZoomOut)

        self.pan()

        self.canvas.setExtent(QgsRectangle(
            crs_transform.transform(QgsPointXY(14.373, 50.034)),
            crs_transform.transform(QgsPointXY(14.500, 50.103)))
        )

        self.canvas.setDestinationCrs(mapy_cz.crs())

        self.data_layers = []

    async def zoomIn(self):
        self.canvas.setMapTool(self.toolZoomIn)

    def zoomOut(self):
        self.canvas.setMapTool(self.toolZoomOut)
        canvas.setMapTool(feature_identifier)

    def pan(self):
        self.canvas.setMapTool(self.toolPan)

    def addLayer(self, layer):
        self.data_layers.append(layer)
        self.pushLayers()

    def removeLayer(self, layer):
        self.data_layers.remove(layer)
        self.pushLayers()

    def pushLayers(self):
        self.canvas.setLayers(list(reversed(self.data_layers)) + [mapy_cz])



# create Qt application
app = QApplication([])


orm  = QgsRasterLayer('crs=EPSG:4326&format&type=xyz&url=https://a.tiles.openrailwaymap.org/standard/%7Bz%7D/%7Bx%7D/%7By%7D.png&zmax=19&zmin=0', 'ORM-standard', 'wms')
osm  = QgsRasterLayer('crs=EPSG:4326&format&type=xyz&url=http://tile.openstreetmap.org/%7Bz%7D/%7Bx%7D/%7By%7D.png&zmax=19&zmin=0', 'ORM-standard', 'wms')
mapy_cz  = QgsRasterLayer('http-header:referer=https://mapy.cz/ &referer=https://mapy.cz/ &type=xyz&url=https://mapserver.mapy.cz/base-m/%7Bz%7D-%7Bx%7D-%7By%7D', 'mapycz-standard', 'wms')

# create main window
wnd = MainWind()
canvas = wnd.map.canvas
wnd.setWindowTitle(f"{window_title_prefix}")
wnd.show()



# run!
event_loop = QEventLoop(app)
asyncio.set_event_loop(event_loop)
app_close_event = asyncio.Event()
app.aboutToQuit.connect(app_close_event.set)

with event_loop:
    event_loop.run_until_complete(app_close_event.wait())
#retval = app.exec_()

# exit
wnd = None
QgsApplication.exitQgis()
sys.exit(0)
