import sys
import time

from qgis.core import *
from qgis.gui import *
from qgis.PyQt.QtXml import QDomDocument, QDomElement
from qgis.PyQt.QtWidgets import *
from qgis.PyQt.QtCore import *
from qgis.PyQt.QtGui import *

import gtfs

from qasync import QEventLoop, asyncClose, asyncSlot, QApplication
import asyncio

import numpy as np

from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import \
    NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure

from utils import *
import communication
from communication import dt_to_path
import data_utils
from data_utils import Trip, TripPoint, HistoryPoint, TripHistory, shape_indexer, dist, unzip_parse


class Graph(QWidget):
    closed = pyqtSignal()

    def closeEvent(self, event):
        self.closed.emit()
        super().closeEvent(event)

    def __init__(self):
        super().__init__()
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)


        canvas = FigureCanvas(Figure(figsize=(5, 3)))
        self._layout.addWidget(canvas)
        self._layout.addWidget(NavigationToolbar(canvas, self))

        self.ax = canvas.figure.subplots()


class HistoryGraph(Graph):
    def __init__(self, stops, default_line=None, default_dt_from=None, default_dt_to=None):
        super().__init__()

        self._add_layout = QHBoxLayout()
        self._layout.insertLayout(0, self._add_layout)

        self._add_selector = QLineEdit(str(default_line) if default_line else "")
        self._add_layout.addWidget(self._add_selector)
        self._add_selector.returnPressed.connect(self.process_add)
        self._add_dt_from = QLineEdit(format_dt_to_user(default_dt_from or datetime.datetime.now(local_timezone)-datetime.timedelta(hours=2)))
        self._add_layout.addWidget(self._add_dt_from)
        self._add_dt_from.returnPressed.connect(self.process_add)
        self._add_dt_to = QLineEdit(format_dt_to_user(default_dt_to or datetime.datetime.now(local_timezone)))
        self._add_layout.addWidget(self._add_dt_to)
        self._add_dt_to.returnPressed.connect(self.process_add)
        self._add_button = QPushButton("Add")
        self._add_layout.addWidget(self._add_button)
        self._add_button.clicked.connect(self.process_add)

        xticks = []
        xlabels = []
        for stop in stops:
            x = stop.shape_dist_traveled
            self.ax.axvline(x=x, ymin=0, ymax=1, color="black") 
            #self.ax.xlabel(xy =(((x-x_bounds[0])/(x_bounds[1]-x_bounds[0])),1.01), xycoords='axes fraction', verticalalignment='top', horizontalalignment='center' , rotation = 270, text=stop.stop.name)
            #self.ax.text(x, 0.5, stop.stop.name, rotation=90, transform=ax.get_xaxis_text1_transform(0)[0])
            #self.ax.text(x, 0.5, stop.stop.name, rotation=90, transform=ax.get_xaxis_text1_transform(0)[0])
            xticks.append(x)
            xlabels.append(stop.stop.name)
        self.ax.set_xticks(xticks)
        self.ax.tick_params(rotation=45)
        self.ax.set_xticklabels(xlabels)

        self.stops_to_shape_dist_traveled = {stop.stop.id: stop.shape_dist_traveled  for stop in stops}


    def add_trip(self, th: TripHistory):
        offsets = [ self.stops_to_shape_dist_traveled[stop.stop.id] - stop.shape_dist_traveled if stop.stop.id in self.stops_to_shape_dist_traveled else None for stop in th.stops ]
        defined_offdets = [ x for x in offsets if x is not None ]
        if len(defined_offdets) == 0:
            print(offsets)
            print("No intersection")
            return
        offset = sorted(defined_offdets)[len(defined_offdets)//2]
        if any(abs(x-offset)  > 0.01 for x in defined_offdets):
            print("Match failed")
            print([ x-offset for x in defined_offdets ])

        x = []
        y = []
        for hp in th.history:
            if hp.shape_point_dist_traveled is not None:
                x.append(hp.shape_point_dist_traveled + offset)
                y.append(hp.first_captured)
        self._line, = self.ax.plot(x, y, 'o-', color="red")

        x = []
        y = []
        for stop in th.stops:
            x += 2*[stop.shape_dist_traveled + offset]
            y += [stop.arrival_time, stop.departure_time]
        self._line, = self.ax.plot(x, y, color="green")

        for stop in th.stops:
            x = 2*[stop.shape_dist_traveled + offset]
            y = [stop.arrival_time, stop.departure_time]
            self.ax.plot(x, y, 'o', color='green' if stop.stop.id in self.stops_to_shape_dist_traveled else 'blue')

        self._line.figure.canvas.draw()

    @asyncSlot()
    async def process_add(self):
        selector = self._add_selector.text()
        dt_from = parse_dt_from_user(self._add_dt_from.text())
        dt_to = parse_dt_from_user(self._add_dt_to.text())

        g = gtfs.for_date(dt_from)
        await g.load_trips()

        trips = [Trip(id, g.date) for id, t in g.trips.items() if id.startswith(selector+"_")]
        ths = {t.trip_id: TripHistory(t) for t in trips}

        route_ids = {trip.trip_id.split("_")[0] for trip in trips}
        print(route_ids)

        c = await get_communication()

        for route_id in route_ids:
            dt = dt_from.replace(second=0, microsecond=0, minute=0)
            while dt < dt_to:
                data, source_timestamps = await c.get_preprocessed_data(dt, route_id)
                print("P1")
                data = await unzip_parse(data)
                print("P3")

                for trip_id, th in ths.items():
                    if trip_id in data:
                        th.add_preprocessed_data(data[trip_id])
                dt += datetime.timedelta(hours=1)

            print("P4")

        for route_id in route_ids:
            stops_data = await c.gtfs_get_stop_times(dt_from, route_filter=route_id)

            for trip_id, th in ths.items():
                if trip_id.split("_")[0] == route_id:
                    await th.load_stops(stops_data)

        for th in ths.values():
            if len(th.history):
                print("ADD TRIP")
                self.add_trip(th)
        print("DONE")

