#!/usr/bin/env pypy3

import asyncio
import datetime
import sys, os
import pathlib
import json
import time
import pprint
import tempfile
import shutil
from pathlib import Path

import data_utils
from data_utils import unzip_parse
import communication
import gtfs
import data_utils
from data_utils import Trip, TripPoint, HistoryPoint, TripHistory, shape_indexer, dist
from utils import *

import gc


async def clean_data_until(ths, dt, loaded_timestamps):
    eprint("START CLEANUP UNTIL", dt)
    y = [x for x in loaded_timestamps if x >= dt]
    loaded_timestamps.clear()
    loaded_timestamps += y
    for i, trip_id in enumerate([x for x in ths]):
        if i%100 == 0:
            await asyncio.sleep(1)
        # Coppy preventing RuntimeError: dictionary changed size during iteration
        if trip_id in ths:
            th = ths[trip_id]
            th.history = [hp for hp in th.history if hp.last_captured >= dt]
            if not th.history:
                eprint(f"REMOVING TRIP {trip_id}")
                del ths[trip_id]
    await asyncio.sleep(1)
    eprint("DONE CLEANUP UNTIL", dt)
    eprint(gc.get_stats())
    gc.collect()
    eprint(gc.get_stats())
    eprint("DONE GC", dt)

def serialize_data(data, dt_from, dt_to):
    out = {}
    for th in data:
        history = [
                {
                    "lat": hp.lat,
                    "lon": hp.lon,
                    "state_position": hp.state_position,
                    "first_captured": hp.first_captured.isoformat(),
                    "last_captured": hp.last_captured.isoformat(),
                    "origin_timestamp": hp.origin_timestamp.isoformat(),
                    "openapi_shape_dist_traveled": hp.openapi_shape_dist_traveled,
                    "openapi_next_stop": hp.openapi_next_stop,
                    "openapi_last_stop": hp.openapi_last_stop,
                    "openapi_delay": hp.openapi_delay,
                    "shape_point": hp.shape_point,
                    "shape_point_dist_traveled": hp.shape_point_dist_traveled,
                }
                for hp in th.history
                if dt_from <= hp.last_captured and hp.first_captured < dt_to
        ]
        if history:
            out[th.trip.trip_id] = {
                'history': history,
                'trip_id': th.trip.trip_id,
                'trip': th.trip_json

        }
    return out

async def write_data(ths, dt, loaded_timestamps):
    assert dt.timestamp() % 3600 == 0

    eprint("PREPARE OUTPUT", dt, Path('data/realtime_by_route')/dt.strftime('%Y-%m-%d')/dt.strftime('%H'))

    source_timestamps = [x for x in loaded_timestamps if dt <= x and x < dt+datetime.timedelta(hours=1)]

    if len(source_timestamps) < 300:
        eprint(f"!!! OUTPUT FAILED there is only {len(source_timestamps)} source timestamps!!!")
        await clean_data_until(ths, dt, loaded_timestamps)
        return

    data_by_route = {}
    for th in ths.values():
        key = th.trip.trip_id.split("_")[0]
        if key not in data_by_route:
            data_by_route[key] = []
        data_by_route[key].append(th)

    tmp_dir = tempfile.mkdtemp(dir="tmp")
    eprint("START OUTPUT", dt, tmp_dir, Path('data/realtime_by_route')/dt.strftime('%Y-%m-%d')/dt.strftime('%H'))

    for route_id, data in data_by_route.items():

        eprint(f"OUTPUT {dt} {route_id}")
        out = serialize_data(data, dt, dt+datetime.timedelta(hours=1))
        if out:
            with open(Path(tmp_dir)/(route_id+".json.zst"), "xb") as f:
                proc = await asyncio.create_subprocess_exec("zstd", "-8", stdout=f, stdin=asyncio.subprocess.PIPE)
                stdout, stderr = await proc.communicate(json.dumps(out).encode("utf-8"))

    with open(Path(tmp_dir)/("source_timestamps"), "x") as f:
        f.write("\n".join([x.isoformat() for x in source_timestamps])+"\n")

    try:
        os.mkdir(Path('data/realtime_by_route')/dt.strftime('%Y-%m-%d'))
    except FileExistsError:
        pass

    out_path = Path('data/realtime_by_route')/dt.strftime('%Y-%m-%d')/dt.strftime('%H')
    if out_path.exists():
        shutil.rmtree(out_path)
    os.rename(tmp_dir, out_path)

    eprint("DONE OUTPUT", dt, tmp_dir, Path('data/realtime_by_route')/dt.strftime('%Y-%m-%d')/dt.strftime('%H'))

    await asyncio.sleep(100)
    await clean_data_until(ths, dt, loaded_timestamps)

async def reload_gtfs(dt):
    eprint("RELOAD GTFS")
    d1 = (dt-datetime.timedelta(hours=4)).date()
    d2 = (dt+datetime.timedelta(hours=4)).date()
    for x in [x for x in gtfs.for_date_cache]:
        if x != d1 and x != d2:
            del gtfs.for_date_cache[x]
    await gtfs.for_date(d1).load_stops_for_all_trips()
    await gtfs.for_date(d2).load_stops_for_all_trips()
    eprint("RELOAD GTFS DONE")


def last_done():
    p1 = Path('data/realtime_by_route')
    for x1 in sorted(os.listdir(p1), reverse=True):
        p2 = p1/x1
        for x2 in sorted(os.listdir(p2), reverse=True):
            return path_to_dt(x1+"-"+x2)

class PreprocessServer(communication.FuncCaller):
    @communication.server_exec()
    async def get_preprocessed_data(self, dt_from: datetime.datetime, dt_to: datetime.datetime, route: str):
        eprint("get_preprocessed_data", dt_from, dt_to, route)
        source_timestamps = [x for x in self.loaded_timestamps if dt_from <= x and x < dt_to]
        out = serialize_data([th for th in self.ths.values() if th.trip.trip_id.split("_")[0] == route], dt_from, dt_to)
        proc = await asyncio.create_subprocess_exec("zstd", "-5", stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE)
        stdout, stderr = await proc.communicate(json.dumps(out).encode("utf-8"))
        eprint("done get_preprocessed_data", dt_from, dt_to, route)
        return stdout, source_timestamps

async def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--dt-from")
    parser.add_argument("--dt-to")
    parser.add_argument("--socket", default="sockets/preprocess_server")
    args = parser.parse_args()

    ths = {}
    loaded_timestamps = []

    async def client_connected(reader, writer):
        s = PreprocessServer(await communication.AsyncioStreamSocket().connect(reader, writer), is_server=True)
        s.ths = ths
        s.loaded_timestamps = loaded_timestamps
        await catch_all(s._server_.run())


    if args.socket != "-":
        await asyncio.start_unix_server(client_connected, path=args.socket, start_serving=True)

    c = await get_communication()

    dt_from = communication.path_to_dt(args.dt_from) if args.dt_from else  last_done()
    dt_to = communication.path_to_dt(args.dt_to) if args.dt_to else None
    # fetch last hour again for better shape matching


    last_flush_dt = dt_from

    dt = dt_from

    await reload_gtfs(dt)
    next_reload_gtfs_dt = dt + datetime.timedelta(hours=1)

    while True:
        if next_reload_gtfs_dt <= dt:
            await reload_gtfs(dt)
            next_reload_gtfs_dt = dt + datetime.timedelta(hours=1)

        r = await c.wait_next_data(dt)
        if r is None:
            eprint("Get next return", r)
            await asyncio.sleep(10)
            continue

        dt = r[0]
        data = await unzip_parse(r[1])


        eprint("PROCESSING", dt)
        data_utils._shape_matching_time = 0
        loaded_timestamps.append(dt)
        _start_time = time.time()
        used_trip_ids = set()
        for dato in data["features"]:
            trip_id = dato["properties"]["trip"]["gtfs"]["trip_id"]
            if trip_id in used_trip_ids:
                eprint("duplicate trip ID", trip_id, flush=False)
            used_trip_ids.add(trip_id)
            if trip_id not in ths:
                ths[trip_id] = TripHistory(Trip(trip_id, datetime.datetime.fromisoformat(dato["properties"]["trip"]["start_timestamp"]).date()))
                await ths[trip_id].load_gtfs_shape()
            ths[trip_id].add_history_point(dt, dato, save_json=False)
        eprint("TIME", time.time() - _start_time, "SHAPE MATCHING", data_utils._shape_matching_time, f"\nSHAPE LOADING: {gtfs._shape_loading_from_cache_count} from disk {gtfs._shape_loading_from_disk_count} from cache\n shape chace sizes { {date.strftime('%Y-%m-%d') :len(g._shape_cache) for date, g in gtfs.for_date_cache.items()} }")
        gtfs._shape_loading_from_disk_count = 0
        gtfs._shape_loading_from_cache_count = 0

        if last_flush_dt+datetime.timedelta(hours=2) <= dt:
            last_flush_dt = last_flush_dt + datetime.timedelta(hours=1)
            loop = asyncio.get_event_loop()
            loop.create_task(catch_all(write_data(ths, last_flush_dt, loaded_timestamps), exitcode=1))

        if dt_to is not None and dt_to <= dt:
            break





if __name__ == "__main__":
    import server
    local_main_server = server.MainServer(is_server=True)
    async def get_communication():
        return local_main_server

    async def gtfs_default_data_getter(date, filename):
        c =  await get_communication()
        return await c.gtfs_get_file(date, filename)

    gtfs.default_data_getter = gtfs_default_data_getter
    data_utils.get_communication = get_communication
    asyncio.run(main())
